<?php
/*******************************************************************************
*	This file is part of Woogle4MediaWiki
*   (http://www.mediawiki.org/wiki/Extension:Woogle4MediaWiki)
*
*	Copyright (c) 2007 - 2009 Hans-Jörg Happel and
*	FZI Forschungszentrum Informatik and der Universität Karlsruhe (TH)
*
*   Woogle4MediaWiki is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   Woogle4MediaWiki is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with Woogle4MediaWiki.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/ 

set_include_path(get_include_path() . PATH_SEPARATOR . dirname(__FILE__));
require_once 'Zend/Search/Lucene.php';

class WoogleLucene {
	
	private static $index;			// woogle zend lucene search index
	private static $indexPath;		// path to woogle zend lucene search index
	
	// TODO (Ben)? to implement: id, uri, label, content, created, changed, size
	
	/**
	 * @return 	String, index directory path
	 */
	public static function getIndexPath() {
		WoogleUtil::profileIn(__METHOD__);
		if (!isset(self::$indexPath)) self::$indexPath = dirname(__FILE__) . '/index';
		return WoogleUtil::profileOut(__METHOD__, self::$indexPath);
	}

	/**
	 * Allows to set a non-standard index papth
	 *
	 * This method is primarily intended to be called by test cases 
	 * 
	 * @param $path		String, index directory path
	 * @return 			void
	 */
	public static function setIndexPath($path){
		WoogleUtil::profileIn(__METHOD__);
		self::$indexPath = $path;
		WoogleUtil::profileOut(__METHOD__);
	}

	/**
	 * Creates instance of a lucene index
	 * If no index exists, a new index will be created
	 * 
	 * @param $overwrite	boolean, if index should be overwritten
	 * @return				Zend_Search_Lucene_Interface 
	 */
	public static function getIndex($overwrite = false) {
		WoogleUtil::profileIn(__METHOD__);
		
		try {
			
			if ($overwrite) {
				// commit previous index if existing
				if (self::$index != null) {
					self::$index->commit();
				}
				// create new index
				self::$index = Zend_Search_Lucene::create(self::getIndexPath());
			
			} else {
				
				// load index if it has not been loaded yet
				if (self::$index == null){
	
					// try if index exists (there does not seem to be an exists() method in current lucene PHP
					try {
						self::$index = Zend_Search_Lucene::open(self::getIndexPath());
					} catch (Zend_Search_Lucene_Exception $e) {
						self::$index = Zend_Search_Lucene::create(self::getIndexPath());
					}
					
				}
			}
			
		} catch (Zend_Search_Lucene_Exception $e) {
			WoogleUtil::error('Error getting index ' . $e->getMessage());
		}
		
		WoogleUtil::profileOut(__METHOD__);
		return self::$index;
	}
	
	/**
	 * Returns number of documents in index
	 * 
	 * @return 		int, number of documents
	 */
	public static function getIndexSize() {
		WoogleUtil::profileIn(__METHOD__);
		try {
			return WoogleUtil::profileOut(__METHOD__, self::getIndex()->numDocs());
		} catch (Zend_Search_Lucene_Exception $e) {
			WoogleUtil::error('WoogleLucene::getIndexSize() ' . print_r($e, true));
			return WoogleUtil::profileOut(__METHOD__, 0);
		}
	}
	
	/**
	 * Create new index for complete re-indexing (i.e. overwrite existing index)
	 * 
	 * @return 	void
	 */
	public static function newIndex() {
		WoogleUtil::profileIn(__METHOD__);
		
		error_reporting( E_ERROR | E_WARNING | E_PARSE );
		// ini_set( 'display_errors', 1 );
		// ini_set('max_execution_time', 900); // now done in WoogleSpecialConfig 
		$index = self::getIndex(true);
		
		WoogleUtil::profileOut(__METHOD__);
	}
	
	/**
	 * Commit index (write changes to disk)
	 * 
	 * @return void
	 */
	public static function commitIndex(){
		WoogleUtil::profileIn(__METHOD__);
		$index = self::getIndex();
		$index->commit();
		WoogleUtil::profileOut(__METHOD__);
	}
		
	/**
	 * Optimize index (may take some time and disk space)
	 * 
	 * @return void
	 */
	public static function optimizeIndex(){
		WoogleUtil::profileIn(__METHOD__);
		$index = self::getIndex();
		$index->optimize();
		WoogleUtil::profileOut(__METHOD__);
	}
	
	/**
	 * @return		float, cumulated size of all files in the index directory in kilobytes
	 */
	public static function getIndexFileSize() {
		WoogleUtil::profileIn(__METHOD__);
		$indexSize = 0;
		$indexDir = opendir(self::getIndexPath());
		while ($file = readdir($indexDir)) {
			$indexSize += filesize(self::getIndexPath() . '/' . $file);
		}
		WoogleUtil::profileOut(__METHOD__);
		return $indexSize / 1024;
	}
	
	/**
	 * Adds a WoogleDocument to the index
	 * 
	 * @param $wd			WoogleDocument
	 * @return 				void
	 */
	public static function addItem($wd) {
		WoogleUtil::profileIn(__METHOD__);
		
		//$content 	= utf8_decode($wd->getContent());	// this is probably required due to MediaWiki
		//$label 		= utf8_decode($wd->getLabel());	// this is probably required due to MediaWiki
		
		$index = self::getIndex();
		Zend_Search_Lucene_Analysis_Analyzer::setDefault( WoogleLucene::getAnalyzer() );
		
		$doc = new Zend_Search_Lucene_Document();

		// core fields
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$id, $wd->getId()));
		//$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$uri, $wd->getUri()));
		$doc->addField(Zend_Search_Lucene_Field::UnIndexed(WoogleFd::$uri, $wd->getUri()));
		$doc->addField(Zend_Search_Lucene_Field::Text(WoogleFd::$label, $wd->getLabel(), 'UTF-8'));
		$doc->addField(Zend_Search_Lucene_Field::Text(WoogleFd::$content, $wd->getContent(), 'UTF-8'));

		// additional fields
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$mimetype, $wd->getMimetype()));
		// MediaWiki concatenated string timestamp (YYYYMMDDHHMMSS). 
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$created, $wd->getCreated() . '000'));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$changed, $wd->getChanged() . '000'));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$source, $wd->getSource()));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$repository, $wd->getRepository()));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$pages, $wd->getPages()));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$size, WoogleUtil::longToString($wd->getSize())));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$sizeChars, WoogleUtil::longToString($wd->getSizeChars())));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$sizeLines, WoogleUtil::longToString($wd->getSizeLines())));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$sizeWords, WoogleUtil::longToString($wd->getSizeWords())));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$scope, $wd->getScope()));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$genre, $wd->getGenre()));
		$doc->addField(Zend_Search_Lucene_Field::Keyword(WoogleFd::$language, $wd->getLanguage()));

		$index->addDocument($doc);
		
		WoogleUtil::profileOut(__METHOD__);
	}
	
	
	/**
	 * Deletes article from the index
	 * 
	 * @param $id			String, identifying the article (=internal MW id?)
	 * @return 				void
	 */
	public static function deleteItem($id) {
		WoogleUtil::profileIn(__METHOD__);
		
		$index = self::getIndex();
		
		$term = new Zend_Search_Lucene_Index_Term(WoogleQueryServiceCommon::getId($id), WoogleFd::$id);
		$query = new Zend_Search_Lucene_Search_Query_Term($term);
		//	$query = Zend_Search_Lucene_Search_QueryParser::parse(WoogleFd::$id . ':"' . WoogleQueryServiceCommon::getId($id) . '"');
		$hits = $index->find($query);
		
		if (count($hits) > 1) {
			$msg = 'More than one article found for id ' . WoogleQueryServiceCommon::getId($id) . '.';
			WoogleUtil::error($msg);
			WoogleUtil::profileOut(__METHOD__);
			throw new Exception($msg);
		} else if (count($hits) == 1) {
			// everything fine - delete article
			$index->delete($hits[0]->id);
			$index->commit();
			WoogleUtil::debug('Article deleted: ' . WoogleFd::$id. ':"' . WoogleQueryServiceCommon::getId($id) . '"');
		} else {
			WoogleUtil::error('Article not found for deletion: ' . WoogleFd::$id . ':"' . WoogleQueryServiceCommon::getId($id) . '"');
		}
		
		WoogleUtil::profileOut(__METHOD__);
	}

	/**
	 *
	 * NOTE: this will return true even for deleted documents, unless optimizeIndex() was called before
	 * 
	 * @param boolean	if a document exists 
	 */
	public function exists($id){
		WoogleUtil::profileIn(__METHOD__);
		$res = false;
		$index = self::getIndex();
		$df = $index->docFreq(new Zend_Search_Lucene_Index_Term(WoogleQueryServiceCommon::getId($id), WoogleFd::$id));
		if ($df > 0) $res = true;
		WoogleUtil::profileOut(__METHOD__);
		return $res; 
	}
	

	public static function getAnalyzer() {
		WoogleUtil::profileIn(__METHOD__);
		// check if PCRE unicode support is available, necessary for UTF-8 indexing
		if ( WoogleLucene::getPcreSupport() ) {
			WoogleUtil::debug( "PCRE unicode support available." );
			// check if the PHP mbstring module is available, that is required for case
			// insensitive indexing and searching 
			if ( WoogleLucene::getMbstringSupport() ) {
				$analyzer = new Zend_Search_Lucene_Analysis_Analyzer_Common_Utf8_CaseInsensitive();
				WoogleUtil::debug( "PHP mbstring module loaded, using caseinsensitive UTF-8 Zend analyzer" );
			} else {
				$analyzer = new Zend_Search_Lucene_Analysis_Analyzer_Common_Utf8();
				WoogleUtil::debug( "PHP mbstring module not loaded, using UTF-8 Zend analyzer" );
			}
		} else {
			$analyzer = Zend_Search_Lucene_Analysis_Analyzer::getDefault();
			WoogleUtil::debug( "PCRE unicode support not available, using default Zend analyzer." );
		}
		return WoogleUtil::profileOut(__METHOD__, $analyzer);
	}


	/**
	 * Checks if the PHP module mbstring is loaded.
	 * 
	 * @return	true if it is loaded
	 */
	public static function getMbstringSupport() {
		WoogleUtil::profileIn(__METHOD__);
		$mbstringLoaded = extension_loaded('mbstring');
		return WoogleUtil::profileOut(__METHOD__, $mbstringLoaded);
	}


	/**
	 * Checks if PCRE unicode support is available. This is needed for
	 * caseinsensitive indexing.
	 * 
	 * @return	true if PCRE available
	 */
	public static function getPcreSupport() {
		WoogleUtil::profileIn(__METHOD__);
		$pcreSupported = false;
		if (@preg_match('/\pL/u', 'a') == 1) {
			$pcreSupported = true;
		}
		return WoogleUtil::profileOut(__METHOD__, $pcreSupported);
	}
}

?>
